/***************************************************************************
 *   Copyright (C) 2004 by Maciej "Kasprzol" Kasprzyk                      *
 *   katalogue@katalogue.szluug.org                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "myplugin.h"
#include <dlfcn.h> // funkcje zwizane z obsug dll'i
#include <iostream>
#include <dirent.h> // struct dirent, opendir(), readdir() & so on...

std::multimap<std::string, MyPlugin*> Rozszerzenia ;
std::vector<MyPlugin*> Pluginy ;

MyPlugin::~MyPlugin ()
{
	if (m_dll_handler)
		dlclose (m_dll_handler) ;
}

int MyPlugin::loadplug (const char * plugin_name)
{
	m_dll_name = plugin_name ;
	m_dll_handler = dlopen (plugin_name, RTLD_LAZY) ;
	if (!m_dll_handler) {
		std::cerr << "Bd podczas adowania biblioteki " << m_dll_name << "!\n" ;
//		perror ("") ;
		m_ok = false ;
		return -1 ;
	} else {
		std::cerr << "Zaadowano bibliotek " << m_dll_name << "\n" ;
	}
	m_get_file_type = (get_file_type_func) dlsym (m_dll_handler, "GetFileType") ;
	m_get_file_info = (get_file_info_func) dlsym (m_dll_handler, "GetFileInfo") ;
	m_display_file_info = (display_file_info_func) dlsym (m_dll_handler, "DisplayFileInfo") ;
	m_get_plugin_info = (get_plugin_info_func) dlsym (m_dll_handler, "GetPluginInfo") ;
	m_destroy_widgets = (destroy_widgets_func) dlsym (m_dll_handler, "DestroyWidgets") ;
	m_is_compatible_with = (is_compatible_with_func) dlsym (m_dll_handler, "IsCompatibleWith") ;
	
	if (!m_get_file_type || !m_get_file_info || !m_display_file_info || !m_get_plugin_info ||
	    !m_destroy_widgets || !m_is_compatible_with) {
		std::cerr << "Bd podczas adowania symboli z biblioteki " << m_dll_name << "\n" ;
		m_ok = false ;
		dlclose (m_dll_handler) ;
		m_dll_handler = NULL ;
		return -1 ;
	}
	m_ok = true ;
	return 0 ;
}

void MyPlugin::initPlugins ()
{
DIR * dir ;
struct dirent * de ;
std::string path = getenv ("HOME") ;
path += (char*)"/.katalogue/plugins/" ;

	dir = opendir (path.c_str ()) ;
	if (dir == NULL) {
		return ;
	}

	while ( (de = readdir (dir)) ) {
		if (strcmp (de->d_name, ".") == 0 || strcmp (de->d_name, "..") == 0) {
			continue ;
		}
		std::string n = path + de->d_name ;
		if (n.substr (n.length () - 3) == ".so") {
			MyPlugin * plug = new MyPlugin (n) ;
			if (plug->loaded ()) {
				Pluginy.push_back (plug) ;
			} else {
				delete plug ;
				continue ;
			}
			// teraz pobierz rozszerzenia obsugiwane przez plugina
			std::string r, rozsz = plug->getFileType () ;
			std::cerr << n << ": Rozszerzenia obsugiwane przez plugin: \"" << rozsz << "\"\n" ;
			unsigned int pb = 0, pe = rozsz.find (";") ;
			do {
				r = rozsz.substr (pb, pe - pb) ;
				std::cerr << "Dodaj rozszerzenie: \"" << r << "\"\n" ;
				Rozszerzenia.insert (make_pair (r, plug)) ;
				if (pe == rozsz.npos) {
					break ;
				}
				pb = pe ;
				pe = rozsz.find (";", ++pb) ;
			} while (true) ;
		}
	}
	closedir (dir) ;
}

void MyPlugin::deInitPlugins ()
{
	std::vector<MyPlugin*>::iterator i, e = Pluginy.end () ;
	for (i = Pluginy.begin () ; i != e ; i++) {
		delete *i ;
	}
}
