/***************************************************************************
 *   Copyright (C) 2004 by Maciej "Kasprzol" Kasprzyk                      *
 *   katalogue@katalogue.szluug.org                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

// nagwki lokalne
#include "treeview.h"
#include "mydirinfo.h"
#include "varibles.h"

// nagwki systemowe c++
#include <map>
#include <string>
#include <iostream> // tymczasowo, dla cerr

// nagwki systemowe C
#include <time.h> // strftime ()
#include <sys/stat.h> // wykorzystywane przez static QSTRing prawa

// nagwki qt
#include <qpixmap.h>
#include <qheader.h>

// nagwki KDE
#include <kmimetype.h>
#include <kurl.h>
#include <klocale.h>
#include <kiconloader.h> // adowanie ikon.

using std::string ;

struct _FileInfo {
	QPixmap* pixmap ;
	QString type ;
	~_FileInfo () { delete pixmap ; }
} ;
typedef std::map<string, _FileInfo*> IconsMap ;
static IconsMap Icons ;

static inline int pow (int base, int exp)
{
	int rc = base ;
	while ( exp-- > 1 ) {
		rc *= base ;
	}
	return rc ;
}

static inline double usun_miejsca_po_przecinku (double in, int ile_zostawic) // napewno inline ?? - wywoywana  6 razy w konstruktorze TreeItem::TreeItem( KListView * parent, BaseClass * bc )
{
	double po_przecinku = in - (int)in ; // konwersja do inta z double ucina cz uamkow bez konwersji
	
	po_przecinku *= pow (10, ile_zostawic) ;
	int b = po_przecinku ;
	return ( (int)in + (double)b / pow (10, ile_zostawic) ) ;
	
}

static QString prawa (uint mode)
{
	QString ret ;
#define MODETEST(TEST,LITERA) if (mode & TEST) {\
		ret += LITERA ;\
	} else {\
		ret += "-" ;\
	}
#define MODETEST2(TEST,LITERA,TEST2,LITERA2) if (mode & TEST) {\
		ret += LITERA ;\
	} else if (mode & TEST2) {\
		ret += LITERA2 ;\
	} else {\
		ret += "-" ;\
	}

	if (S_ISDIR (mode)) {
		ret = "d" ;
	} else if (S_ISCHR (mode)) {
		ret = "c" ;
	} else if (S_ISBLK (mode)) {
		ret = "b" ;
	} else if (S_ISLNK (mode)) {
		ret = "l" ;
	} else {
		ret = "-" ;
	}
	MODETEST (S_IRUSR, "r")
	MODETEST (S_IWUSR, "w")
	MODETEST2 (S_IXUSR, "x", S_ISUID, "S")
	
	MODETEST (S_IRGRP, "r")
	MODETEST (S_IWGRP, "w")
	MODETEST2 (S_IXGRP, "x", S_ISGID, "S")
	
	MODETEST (S_IROTH, "r")
	MODETEST (S_IWOTH, "w")
	MODETEST2 (S_IXOTH, "x", S_ISVTX, "T")
	
	return ret ;
}

/*****************************************************************************
 *
 * Class TreeItem
 *
 *****************************************************************************/

TreeItem::TreeItem( TreeItem * parent, const QString& new_title, BaseClass * bc )
    : KListViewItem( parent, new_title ), item(bc), title(new_title),
      parent_treeitem (parent), parent_listview (NULL), pix( 0 )
{
	// zezwalaj na zmiane nazw gwnego katalogu, grup i poszczeglnych zeskanowanych katalogw (ale tylko top-level katalogu)
	if (bc->type >= BaseClass::DiskGroup || bc->type == BaseClass::TopLevelDir) {
		setRenameEnabled (0, true) ;
	}
}

TreeItem::TreeItem( KListView * parent, const QString& new_title, BaseClass * bc )
    : KListViewItem( parent, new_title ), item(bc), title(new_title),
      parent_treeitem (NULL), parent_listview (parent), pix( 0 )
{
	// zezwalaj na zmiane nazw gwnego katalogu, grup i poszczeglnych zeskanowanych katalogw (ale tylko top-level katalogu)
	if (bc->type >= BaseClass::DiskGroup || bc->type == BaseClass::TopLevelDir) {
		setRenameEnabled (0, true) ;
	}
}
/*
TreeItem::TreeItem( TreeItem * parent, BaseClass * bc )
	: QListViewItem ( parent ), item (bc), title (bc->name.c_str ()), parent_treitem (parent), pix (0)
{
	assert (bc->type == BaseClass::File) ;
	if (bc->type == BaseClass::File) {
		QString str ;
		MyFileInfo * mfi = static_cast<MyFileInfo*>(bc) ;
		struct tm t ;
		char tt[20] ;
		str = mfi->name.c_str () ;
		this->setText (0, str) ;
		str = QString ("%1").arg (mfi->size) ;
		this->setText (1, str) ;
		t = *localtime (&mfi->mtime) ;
		strftime (tt, 19, "%d/%m/%Y %R", &t) ; // %R == %H:%M
		setText (2, tt) ;
		str = QString ("%1").arg (mfi->mode) ;
		setText (3, str) ;
		str = QString ("%1/%2").arg (mfi->uid).arg (mfi->gid) ;
		setText (4, str) ;
	}
}
*/
TreeItem::TreeItem( KListView * parent, BaseClass * bc )
	: KListViewItem ( parent ), item (bc), title (bc->name.c_str ()), parent_treeitem (NULL), parent_listview (parent), pix (0)

{
//	assert (bc->type == BaseClass::File || bc->type == BaseClass::Dir || bc->type == BaseClass::DiskGroup) ;
	switch (bc->type) {
		case BaseClass::LinkToFile:
			setText (5, "Link to file") ; // FIXME: i tak zostanie zmienione pniej
		case BaseClass::File: {
			// pobierz ikon uywajac mime
			string s ;
			
			//std::cout << "Pozycja rozszerzenia: " << bc->name.rfind (".") << "\t";
			if (bc->name.rfind (".") == string::npos) {
				s = bc->name ;
			} else {
				s.assign (bc->name, bc->name.rfind ("."), string::npos) ;
			}
			//std::cout << "rozszerzenie pliku " << s << "\n" ;
			if (Icons[s]) {
				setPixmap (Icons[s]->pixmap) ;
				setText (5, Icons[s]->type) ;
			} else {
				_FileInfo * f ;
				QString bla = "/sadadsadsasasadaddadaxcxzvvddfwfwf/dsadsadsadsadsa/" ;  // jaki katalog ktry na 99.999% nie
				bla += bc->name.c_str () ;						// istnieje u usera na dysku
			
				KURL k (bla) ;
				KMimeType::Ptr type = KMimeType::findByURL(k);
				if (type->name() == KMimeType::defaultMimeType()) {
//					std::cerr << "Could not find out type for file " << bc->name << "\n" ;
				}
				//std::cout << "Type: " << type->name() << "\t" ; 
				type->pixmap (KIcon::NoGroup, KIcon::SizeSmall, 0, &bla) ;
				f = new _FileInfo ;
				f->pixmap = new QPixmap (bla) ;
				f->type = type->name () ;
				setPixmap (f->pixmap) ;
				setText (5, f->type) ;
				Icons[s] = f ;
			}
			//koniec mime
		} // nie break'owa !!!
		case BaseClass::Dir: 
		case BaseClass::CompressedDir:
		case BaseClass::TopLevelDir: {
			QString str ;
			MyFileInfo * mfi = static_cast<MyFileInfo*>(bc) ;
			struct tm t ;
			char tt[20] ;
			
			if (bc->type >= BaseClass::Dir) {
				setPixmap (Icons["inode/directory"]->pixmap) ;
				setText (5, Icons["inode/directory"]->type) ;
			}
			this->setText (0, QString::fromLocal8Bit (mfi->name.c_str ())) ;
			if (mfi->size < 1024) { // < 1 KB
				long a = mfi->size ; // konwersja z maego unsigned long long do long
				str = QString ("%1 B").arg (a) ;
			} else if (mfi->size < 1024 * 1024) { // < 1 MB
				str = QString ("%1 KB").arg (usun_miejsca_po_przecinku (mfi->size / 1024.0, 2)) ; // 1024 jako double poto aby wynikiem dzielenia byo double, a nie int konwertowany do double
			} else if (mfi->size < 1024 * 1024 * 1024) { // < 1 GB :)
				str = QString ("%1 MB").arg (usun_miejsca_po_przecinku (mfi->size / (1024.0 * 1024.0), 2)) ;
			} else {
				str = QString ("%1 GB").arg (usun_miejsca_po_przecinku (mfi->size / (1024.0 * 1024.0 * 1024.0), 2)) ;
//				cout << str ;
			}
			this->setText (1, str) ;
			t = *localtime (&mfi->mtime) ;
			strftime (tt, 19, "%d/%m/%Y %H:%M", &t) ; // warning: ISO C++ does not support the `%R' strftime format (%R == %H:%M)
			setText (2, tt) ;
			// str = QString ("%1").arg (mfi->mode) ;
			str = prawa (mfi->mode) ;
			setText (3, str) ;
			str = QString ("%1/%2").arg (mfi->uid).arg (mfi->gid) ;
			setText (4, str) ;
			if (parent->columns () == 7) { // jeli parent to lista z okna szukania plikw to ma dodatkow kolumn z pen ciek
				setText (6, bc->AbsKataloguePath ().c_str ()) ;
			}
			break ;
		}
		case BaseClass::DiskGroup: {
			QString str ;
			MyDiskGroup * mdg = static_cast<MyDiskGroup*>(bc) ;
			unsigned long long s = mdg->size () ;
			
			setPixmap (Icons["group"]->pixmap) ;
			setText (5, i18n ("group")) ;
			str = mdg->name.c_str () ;
			this->setText (0, str) ;
			if (s < 1024) { // < 1 KB
				long a = s ; // konwersja z maego unsigned long long do long
				str = QString ("%1").arg (a) ;
			} else if (s < 1024 * 1024) { // < 1 MB
				str = QString ("%1 KB").arg (usun_miejsca_po_przecinku (s / 1024.0, 2)) ;
			} else if (s < 1024 * 1024 * 1024) { // < 1 GB :)
				str = QString ("%1 MB").arg (usun_miejsca_po_przecinku (s / (1024.0 * 1024.0), 2)) ;
			} else {
				str = QString ("%1 GB").arg (usun_miejsca_po_przecinku (s / (1024.0 * 1024.0 * 1024.0), 2)) ;
//				cout << str ;
			}
			this->setText (1, str) ;
			break ;
		}
/*		default:
			std::cerr << "Zy argument - typ: " << bc->type << ". " __FILE__ "@" << __LINE__ << "\n" ;
			break ;
*/	}
}

void TreeItem::setPixmap( QPixmap *px )
{
    pix = px;
    setup();
    widthChanged( 0 );
    invalidateHeight();
    repaint();
}

const QPixmap *TreeItem::pixmap( int i ) const
{
    if ( i )
	return 0;
    return pix;
}

void TreeItem::setOpen( bool o )
{
	if (item->type == BaseClass::DiskGroup) {
		setPixmap ( Icons["group"]->pixmap ) ;
	} else {
		if ( o )
			setPixmap( Icons["inode/directory-open"]->pixmap );
		else
			setPixmap( Icons["inode/directory"]->pixmap );
	}

    QListViewItem::setOpen( o );
}

/*
QString TreeItem::text () const
{
	return title ;
}*/

int TreeItem::compare (QListViewItem * qlvi, int col, bool ascending) const
{
/*
Compares this listview item to i using the column col in ascending order. Returns -1 if this item is less than i,
0 if they are equal and 1 if this item is greater than i. 
*/
	TreeItem * i = static_cast<TreeItem*>(qlvi) ;
	if (Varibles::DirectoriesFirst) {
		bool FirstIsDir = false, SecondIsDir = false ;
		if (getItem ()->type >= BaseClass::Dir && getItem ()->type <= BaseClass::TopLevelDir)
			FirstIsDir = true ;
		if (i->getItem ()->type >= BaseClass::Dir && i->getItem ()->type <= BaseClass::TopLevelDir)
			SecondIsDir = true ;
		if (ascending) {
			if (FirstIsDir && !SecondIsDir) return -1 ;
			else if (!FirstIsDir && SecondIsDir) return 1 ;
		} else {
			if (FirstIsDir && !SecondIsDir) return 1 ;
			else if (!FirstIsDir && SecondIsDir) return -1 ;
		}
		// else albo oba s katalogami albo oba nie s wic sortujemy wedug ustalonego porzdku.
	}
	switch (col) {
	case 1: {// rozmiar
		//  cholera, naraz moe by sortowane i MyDirInfo i MyDiskGroup. W MyDirInfo size jest zmienn, a w MyDiskGroup metod :-(
		unsigned long long s1, s2 ;
		if (getItem ()->type >= BaseClass::File && getItem ()->type <= BaseClass::CompressedDir)  { // klasy dziedziczce z MyFileInfo
			s1 = static_cast<MyDirInfo*>(getItem ())->size ;
		} else /* if (getItem ()->type == BaseClass::DiskGroup) */ {
			s1 = static_cast<MyDiskGroup*>(getItem ())->size () ;
		} // else = bd
		if (i->getItem ()->type >= BaseClass::File && i->getItem ()->type <= BaseClass::CompressedDir)  { // klasy dziedziczce z MyFileInfo
			s2 = static_cast<MyDirInfo*>(i->getItem ())->size ;
		} else /* if (i->getItem ()->type == BaseClass::DiskGroup) */ {
			s2 = static_cast<MyDiskGroup*>(i->getItem ())->size () ;
		}
//		std::cerr << "porwnuje rozmiar: " << s1 << " z " << s2 << "\n" ;
		if (s1 < s2) return -1 ;
		else if (s1 > s2) return 1 ;
		else return 0 ;
	}
	case 2: {// data - wedug daty mona sortowa tylko klasy dziedziczce z MyFileInfo
		if (getItem ()->type == BaseClass::DiskGroup) return -1 ;
		if (i->getItem ()->type == BaseClass::DiskGroup) return 1 ;
		return static_cast<MyFileInfo*>(getItem ())->mtime - static_cast<MyFileInfo*>(i->getItem ())->mtime ;
	}
	default: // nazwa pliku, uid, typ
		return key( col, ascending ).compare( i->key( col, ascending) );
	}
}

/*****************************************************************************
 *
 * Class TreeView
 *
 *****************************************************************************/

TreeView::TreeView( QWidget *parent, const char *name)
    : KListView( parent, name )/*, oldCurrent( 0 ),
      dropItem( 0 ), mousePressed( FALSE )*/
{
//    autoopen_timer = new QTimer( this );
    connect( this, SIGNAL( executed( QListViewItem * ) ),
	     this, SLOT( slotTreeItemSelected( QListViewItem * ) ) );
    connect( this, SIGNAL( returnPressed( QListViewItem * ) ),
	     this, SLOT( slotTreeItemSelected( QListViewItem * ) ) );
    connect( this, SIGNAL( itemRenamed( QListViewItem *, int, const QString & ) ),
	     this, SLOT( slotItemRenamed( QListViewItem *, int, const QString & ) ) );

/*
    setAcceptDrops( TRUE );
    viewport()->setAcceptDrops( TRUE );

    connect( autoopen_timer, SIGNAL( timeout() ),
	     this, SLOT( openFolder() ) );
*/
}

void TreeView::initIconCache () // funkcja statyczna
{
	// pobierz ikon uywajac mime
	string mime = "inode/directory" ;
	_FileInfo * f ;
	QString bla  ;

	// znajd ikon przypisan do typu inode/directory
	KMimeType::Ptr type = KMimeType::mimeType(mime.c_str ()) ;
	type->pixmap (KIcon::NoGroup, KIcon::SizeSmall, 0, &bla) ;
	f = new _FileInfo ;
	f->pixmap = new QPixmap (bla) ;
	f->type = i18n ("directory") ;
	//cout << bla << "\n" ;
	Icons[mime] = f ;

	// a teraz ikona folderOpen
	const char * ___tmp = bla ; // FIXME: "sprytna" konwersja z QString'a na string'a
	string s = ___tmp ;

	string tmp1 (s, 0, s.rfind (".")) ;
	//std::cout << tmp1 << "\n" ;
	tmp1 += "_open" ;
	tmp1.append (s, s.rfind ("."), string::npos) ;
	//std::cout << tmp1 << "\n" ;

	f = new _FileInfo ;
	if ( (f->pixmap = new QPixmap (tmp1.c_str ())) ) { // warning: suggest parentheses around assignment used as truth value

		f->type = "" ;
		Icons["inode/directory-open"] = f ;
	} else {
		delete f ;
		std::cerr << "Nie znaleziono ikony dla otwartego katalogu (" << tmp1 << ")\n" ;
		Icons["inode/directory-open"] = Icons[mime] ;
	}

	KIconLoader * kil = KGlobal::iconLoader () ;
	f = new _FileInfo ;
	f->type = "" ;
	f->pixmap = new QPixmap (kil->loadIcon ("kpackage", KIcon::MainToolbar)) ;
	Icons["group"] = f ;
}

void TreeView::deInitIconCache () // funkcja statyczna
{
	IconsMap::iterator i, e = Icons.end () ;
	for (i = Icons.begin () ; i != e ; i++) {
		delete i->second ;
	}
}

void TreeView::slotTreeItemSelected( QListViewItem * i)
{
    if ( !i )
	return ;

    TreeItem *dir = (TreeItem*)i ;
//    emit TreeItemSelected ( dir->text() ) ;
    emit TreeItemSelected ( *dir ) ;
}

void TreeView::slotItemRenamed( QListViewItem * item, int /*col*/, const QString & text )
{
//	static_cast<TreeItem*>(item)->title = text ;
	// trzeba emitowa sygna gdy po zmianie nazwy trzeba ustawi flag modyfikacji
	emit itemRenamed (static_cast<TreeItem*>(item), text) ;
}

void TreeView::setDir (const BaseClass * bc)
{
	TreeItem * it = static_cast<TreeItem*>(currentItem ()) ; // (selectedItem ()) ;//
	if (!it) {
		std::cerr << __FILE__ << "@" << __LINE__ << " : it == NULL!!!\n" ;
		return ;
	}
	it->setOpen (true) ;
//	it->setSelected (false) ;
	TreeItem * child = static_cast<TreeItem*>(it->firstChild ()) ;
//	std::cerr << "Ojciec: " << it->getItem ()->name << "\n" ;
//	std::cerr << "dziecko: " << bc->name << "\n" ;
	while (child != NULL) {
		if (child->getItem () == bc) {
			child->setOpen (true) ;
			child->setSelected (true) ;
			/*this->*/ setCurrentItem (child) ;
//			std::cerr << "Znaleziono dziecko\n" ;
			return ;
		}
		child = static_cast<TreeItem*>(child->nextSibling ()) ;
	}
//	std::cerr << "Nie znaleziono dziecka\n" ;
}

/* Zakomentowane rzeczy zwizane z drag&drop


void TreeView::openFolder()
{
    autoopen_timer->stop();
    if ( dropItem && !dropItem->isOpen() ) {
	dropItem->setOpen( TRUE );
	dropItem->repaint();
    }
}

static const int autoopenTime = 750;


void TreeView::contentsDragEnterEvent( QDragEnterEvent *e )
{
    if ( !QUriDrag::canDecode(e) ) {
	e->ignore();
	return;
    }

    oldCurrent = currentItem();

    QListViewItem *i = itemAt( contentsToViewport(e->pos()) );
    if ( i ) {
	dropItem = i;
	autoopen_timer->start( autoopenTime );
    }
}


void TreeView::contentsDragMoveEvent( QDragMoveEvent *e )
{
    if ( !QUriDrag::canDecode(e) ) {
	e->ignore();
	return;
    }

    QPoint vp = contentsToViewport( ( (QDragMoveEvent*)e )->pos() );
    QListViewItem *i = itemAt( vp );
    if ( i ) {
	setSelected( i, TRUE );
	e->accept();
	if ( i != dropItem ) {
	    autoopen_timer->stop();
	    dropItem = i;
	    autoopen_timer->start( autoopenTime );
	}
	switch ( e->action() ) {
	case QDropEvent::Copy:
	    break;
	case QDropEvent::Move:
	    e->acceptAction();
	    break;
	case QDropEvent::Link:
	    e->acceptAction();
	    break;
	default:
	    ;
	}
    } else {
	e->ignore();
	autoopen_timer->stop();
	dropItem = 0;
    }
}

void TreeView::contentsDragLeaveEvent( QDragLeaveEvent * )
{
    autoopen_timer->stop();
    dropItem = 0;

    setCurrentItem( oldCurrent );
    setSelected( oldCurrent, TRUE );
}

void TreeView::contentsDropEvent( QDropEvent *e )
{
    autoopen_timer->stop();

    if ( !QUriDrag::canDecode(e) ) {
	e->ignore();
	return;
    }

    QListViewItem *item = itemAt( contentsToViewport(e->pos()) );
    if ( item ) {

	QStrList lst;

	QUriDrag::decode( e, lst );

	QString str;

	switch ( e->action() ) {
	    case QDropEvent::Copy:
	    str = "Copy";
	    break;
	    case QDropEvent::Move:
	    str = "Move";
	    e->acceptAction();
	    break;
	    case QDropEvent::Link:
	    str = "Link";
	    e->acceptAction();
	    break;
	    default:
	    str = "Unknown";
	}

	str += "\n\n";

	e->accept();

	for ( uint i = 0; i < lst.count(); ++i ) {
	    QString filename = lst.at( i );
	    str += filename + "\n";
	}
	str += QString( "\nTo\n\n   %1" )
	       .arg( fullPath(item) );

	QMessageBox::information( this, "Drop target", str, "Not implemented" );
    } else
	e->ignore();

}


QString TreeView::fullPath(QListViewItem* item)
{
    QString fullpath = item->text(0);
    while ( (item=item->parent()) ) {
	if ( item->parent() )
	    fullpath = item->text(0) + "/" + fullpath;
	else
	    fullpath = item->text(0) + fullpath;
    }
    return fullpath;
}

void TreeView::contentsMousePressEvent( QMouseEvent* e )
{
    QListView::contentsMousePressEvent(e);
    QPoint p( contentsToViewport( e->pos() ) );
    QListViewItem *i = itemAt( p );
    if ( i ) {
	// if the user clicked into the root decoration of the item, don't try to start a drag!
	if ( p.x() > header()->cellPos( header()->mapToActual( 0 ) ) +
	     treeStepSize() * ( i->depth() + ( rootIsDecorated() ? 1 : 0) ) + itemMargin() ||
	     p.x() < header()->cellPos( header()->mapToActual( 0 ) ) ) {
	    presspos = e->pos();
	    mousePressed = TRUE;
	}
    }
}

void TreeView::contentsMouseMoveEvent( QMouseEvent* e )
{
    if ( mousePressed && ( presspos - e->pos() ).manhattanLength() > QApplication::startDragDistance() ) {
	mousePressed = FALSE;
	QListViewItem *item = itemAt( contentsToViewport(presspos) );
	if ( item ) {
	    QString source = fullPath(item);
	    if ( QFile::exists(source) ) {
		QUriDrag* ud = new QUriDrag(viewport());
		ud->setUnicodeUris( source );
		if ( ud->drag() )
		    QMessageBox::information( this, "Drag source",
					      QString("Delete ")+source, "Not implemented" );
	    }
	}
    }
}

void TreeView::contentsMouseReleaseEvent( QMouseEvent * )
{
    mousePressed = FALSE;
}
*/

